import logging

from asphalt.core.context import Context
import pytest

from asphalt.mailer.api import Mailer
from asphalt.mailer.component import MailerComponent


@pytest.mark.parametrize('backend', ['smtp', 'sendmail'])
@pytest.mark.asyncio
async def test_component_single(caplog, backend):
    caplog.set_level(logging.INFO, logger='asphalt.mailer.component')
    component = MailerComponent(backend=backend)
    async with Context() as ctx:
        await component.start(ctx)
        assert isinstance(ctx.mailer, Mailer)

    records = [record for record in caplog.records if record.name == 'asphalt.mailer.component']
    records.sort(key=lambda r: r.message)
    assert len(records) == 2
    assert records[0].message.startswith(
        "Configured mailer (default / ctx.mailer; class=%s)" % ctx.mailer.__class__.__name__)
    assert records[1].message.startswith('Mailer (default) stopped')


@pytest.mark.asyncio
async def test_component_multiple(caplog):
    caplog.set_level(logging.INFO, logger='asphalt.mailer.component')
    component = MailerComponent(mailers={
        'smtp': {'backend': 'smtp', 'context_attr': 'mailer1'},
        'sendmail': {'backend': 'sendmail', 'context_attr': 'mailer2'}
    })
    async with Context() as ctx:
        await component.start(ctx)
        assert isinstance(ctx.mailer1, Mailer)
        assert isinstance(ctx.mailer2, Mailer)

    records = [record for record in caplog.records if record.name == 'asphalt.mailer.component']
    records.sort(key=lambda r: r.message)
    assert len(records) == 4
    assert records[0].message.startswith(
        "Configured mailer (sendmail / ctx.mailer2; class=SendmailMailer)")
    assert records[1].message.startswith(
        "Configured mailer (smtp / ctx.mailer1; class=SMTPMailer)")
    assert records[2].message.startswith('Mailer (sendmail) stopped')
    assert records[3].message.startswith('Mailer (smtp) stopped')



import logging
from email.message import EmailMessage
from ssl import SSLContext
from typing import Iterable, Union, Dict, Any

from aiosmtplib import SMTP, SMTPTimeoutError
from asphalt.core import Context
from typeguard import check_argument_types

from asphalt.mailer.api import Mailer, DeliveryError

logger = logging.getLogger(__name__)


class SMTPMailer(Mailer):
    """
    A mailer that uses `aiosmtplib`_ to send mails.

    The default port is chosen as follows:

    * 587: if ``username`` and ``password`` have been defined and ``tls`` is ``True``
    * 25: in all other cases

    :param host: host name of the SMTP server
    :param port: override the default port (see above)
    :param tls: whether to initiate TLS using STARTTLS once connected (defaults to ``True`` if
        ``username`` and ``password`` have been defined)
    :param tls_context: either an :class:`~ssl.SSLContext` instance or the resource name of one
    :param username: user name to authenticate as
    :param password: password to authenticate with
    :param timeout: timeout (in seconds) for all network operations
    :param message_defaults: default values for omitted keyword arguments of
        :meth:`~asphalt.mailer.api.Mailer.create_message`

    .. _aiosmtplib: https://github.com/cole/aiosmtplib
    """

    def __init__(self, *, host: str = 'localhost', port: int = None, tls: bool = None,
                 tls_context: Union[str, SSLContext] = None,
                 username: str = None, password: str = None, timeout: float = 10,
                 message_defaults: Dict[str, Any] = None):
        assert check_argument_types()
        super().__init__(message_defaults or {})
        self.host = host
        self.tls = tls if tls is not None else bool(username and password)
        self.port = port or (587 if username and password and self.tls else 25)
        self.tls_context = tls_context
        self.username = username
        self.password = password
        self.timeout = timeout
        self._smtp = None  # type: SMTP

    async def start(self, ctx: Context):
        if isinstance(self.tls_context, str):
            self.tls_context = await ctx.request_resource(SSLContext, self.tls_context)

        self._smtp = SMTP(hostname=self.host, port=self.port, tls_context=self.tls_context,
                          loop=ctx.loop, timeout=self.timeout)
        ctx.add_teardown_callback(self._smtp.close)

    async def deliver(self, messages: Union[EmailMessage, Iterable[EmailMessage]]):
        assert check_argument_types()
        if isinstance(messages, EmailMessage):
            messages = [messages]

        try:
            try:
                await self._smtp.connect()

                # Switch to TLS if required
                if self.tls and not self._smtp.use_tls:
                    await self._smtp.starttls()

                # Authenticate if needed
                if None not in (self.username, self.password):
                    await self._smtp.login(self.username, self.password)
            except Exception as e:
                raise DeliveryError(str(e)) from e

            for message in messages:
                try:
                    await self._smtp.send_message(message)
                except Exception as e:
                    raise DeliveryError(str(e), message) from e
        finally:
            if self._smtp.is_connected:
                try:
                    await self._smtp.quit()
                except (ConnectionError, SMTPTimeoutError):  # pragma: nocover
                    self._smtp.close()

    def __repr__(self):
        return '{self.__class__.__name__}(host={self.host!r}, port={self.port})'.format(self=self)



import ssl
from base64 import b64decode
from contextlib import suppress
from pathlib import Path
from types import MethodType

import pytest
from aiosmtpd.handlers import Message
from aiosmtpd.smtp import SMTP
from asphalt.core.context import Context

from asphalt.mailer.api import DeliveryError
from asphalt.mailer.mailers.smtp import SMTPMailer

try:
    from asyncio import all_tasks
except ImportError:
    def all_tasks(loop=None):
        from asyncio import Task
        return {t for t in Task.all_tasks(loop) if not t.done()}


@pytest.fixture
def handler():
    class MessageHandler(Message):
        def __init__(self, message_class=None):
            super().__init__(message_class)
            self.messages = []

        def handle_message(self, message):
            self.messages.append(message)

    return MessageHandler()


@pytest.fixture(scope='module')
def client_tls_context():
    cert_path = Path(__file__).with_name('server.cert')
    return ssl.create_default_context(cafile=str(cert_path))


@pytest.fixture(scope='module')
def server_tls_context():
    key_path = Path(__file__).with_name('server.key')
    cert_path = Path(__file__).with_name('server.cert')
    ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ctx.load_cert_chain(str(cert_path), str(key_path))
    return ctx


@pytest.fixture(params=[False, True], ids=['plaintext', 'tls'])
def tls(request):
    return request.param


@pytest.fixture
def smtp_server(event_loop, unused_tcp_port, handler, server_tls_context, tls):
    smtp = SMTP(handler, require_starttls=tls, tls_context=server_tls_context)
    server = event_loop.run_until_complete(
        event_loop.create_server(lambda: smtp, port=unused_tcp_port))
    yield smtp
    server.close()
    event_loop.run_until_complete(server.wait_closed())

    # The server leaves tasks running so cancel them
    for task in all_tasks(event_loop):
        task.cancel()
        with suppress(Exception):
            event_loop.run_until_complete(task)


@pytest.fixture
def mailer(event_loop, unused_tcp_port, client_tls_context, smtp_server, tls):
    mailer = SMTPMailer(port=unused_tcp_port, timeout=1, tls=tls, tls_context=client_tls_context)
    with Context() as ctx:
        event_loop.run_until_complete(mailer.start(ctx))
        yield mailer


@pytest.mark.parametrize('username, password, tls, expected_port', [
    ('foo', 'bar', True, 587),
    ('foo', 'bar', None, 587),
    ('foo', 'bar', False, 25),
    ('foo', None, None, 25)
], ids=['forcetls', 'autotls', 'disabletls', 'nopassword'])
def test_port_selection(username, password, tls, expected_port):
    mailer = SMTPMailer(username=username, password=password, tls=tls)
    assert mailer.port == expected_port


@pytest.mark.asyncio
async def test_resources():
    mailer = SMTPMailer(tls_context='contextresource')
    context = Context()
    sslcontext = ssl.create_default_context()
    context.add_resource(sslcontext, 'contextresource')
    await mailer.start(context)
    assert mailer.tls_context is sslcontext


@pytest.mark.asyncio
async def test_deliver(mailer, sample_message, handler):
    await mailer.deliver(sample_message)
    assert len(handler.messages) == 1
    received_message = handler.messages[0]

    headers = dict(received_message.items())
    del headers['X-Peer']
    assert headers == {
        'From': 'foo@bar.baz',
        'To': 'Test Recipient <test@domain.country>, test2@domain.country',
        'Cc': 'Test CC <testcc@domain.country>, testcc2@domain.country',
        'Content-Type': 'text/plain; charset="utf-8"',
        'Content-Transfer-Encoding': '7bit',
        'MIME-Version': '1.0',
        'X-MailFrom': 'foo@bar.baz',
        'X-RcptTo': ('test@domain.country, test2@domain.country, testcc@domain.country, '
                     'testcc2@domain.country, testbcc@domain.country, testbcc2@domain.country')
    }
    assert received_message.get_payload() == 'Test content'


@pytest.mark.asyncio
async def test_deliver_auth(mailer, sample_message, smtp_server, handler):
    """Test that authentication works."""
    async def handle_EHLO(self, smtp, session, envelope, hostname):
        session.host_name = hostname
        return '250 AUTH PLAIN'

    async def smtp_AUTH(self, arg):
        method, credentials = arg.split(' ')
        credentials = b64decode(credentials)
        if method == 'PLAIN' and credentials == b'\x00foo\x00bar':
            await self.push('235 2.7.0 Authentication successful')
        else:
            await self.push('535 5.7.8 Authentication credentials invalid')

    handler.handle_EHLO = MethodType(handle_EHLO, handler)
    smtp_server.smtp_AUTH = MethodType(smtp_AUTH, smtp_server)
    mailer.username = 'foo'
    mailer.password = 'bar'
    await mailer.deliver(sample_message)


@pytest.mark.asyncio
async def test_deliver_connect_error(mailer, sample_message, unused_tcp_port):
    mailer._smtp.port = unused_tcp_port + 1
    with pytest.raises(DeliveryError) as exc:
        await mailer.deliver(sample_message)

    exc.match('Error connecting to localhost on port %d' % mailer._smtp.port)


@pytest.mark.asyncio
async def test_deliver_error(mailer, sample_message, smtp_server):
    """Test that SMTP errors during message delivery get converted into DeliveryErrors."""
    async def smtp_DATA(self, arg):
        await self.push('503 Error: foo')

    smtp_server.smtp_DATA = MethodType(smtp_DATA, smtp_server)
    with pytest.raises(DeliveryError) as exc:
        await mailer.deliver(sample_message)

    exc.match('Error: foo')


def test_repr(mailer, unused_tcp_port):
    assert repr(mailer) == "SMTPMailer(host='localhost', port=%d)" % unused_tcp_port



"""
A simple command line tool that connects to an SMTP server, sends a mail message with an attachment
and then exits.
"""

import logging

import click
from asphalt.core import CLIApplicationComponent, Context, run_application


class ApplicationComponent(CLIApplicationComponent):
    def __init__(self, host, username, password, sender, to, subject, attachment):
        super().__init__()
        self.host = host
        self.username = username
        self.password = password
        self.sender = sender
        self.to = to
        self.subject = subject
        self.attachment = attachment

    async def start(self, ctx: Context):
        self.add_component('mailer', backend='smtp', host=self.host, username=self.username,
                           password=self.password)
        await super().start(ctx)

    async def run(self, ctx: Context):
        message = ctx.mailer.create_message(
            subject=self.subject, sender=self.sender, to=self.to,
            plain_body='Take a look at this file!')
        await ctx.mailer.add_file_attachment(message, self.attachment)
        await ctx.mailer.deliver(message)


@click.command()
@click.option('-h', '--host', required=True, help='SMTP server host name')
@click.option('-u', '--username', help='username for authenticating with SMTP server')
@click.option('-p', '--password', help='password for authenticating with SMTP server')
@click.option('-f', '--from', 'sender', required=True, help='Sender email address')
@click.option('-t', '--to', required=True, help='Recipient email address')
@click.option('-s', '--subject', default='Test email with attachment')
@click.argument('attachment', type=click.Path(exists=True))
def main(host, username, password, sender, to, subject, attachment):
    component = ApplicationComponent(host, username, password, sender, to, subject, attachment)
    run_application(component, logging=logging.INFO)

main()
